# built-in
import os
from contextlib import contextmanager


@contextmanager
def chdir(path):
    """Context manager for changing dir and restoring previous workdir after exit.
    """
    curdir = os.getcwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(curdir)



# built-in
from pathlib import Path

# app
from .._constants import ExitCode
from .._logic import YesQA
from .._types import CommandResult


def get_paths(paths):
    for path in paths:
        if path.is_dir():
            yield from get_paths(path.iterdir())
            continue
        if path.suffix != '.py':
            continue
        if not path.is_file():
            continue
        yield path


def yesqa_command(argv) -> CommandResult:
    """Remove bare and unused noqa comments.
    """
    if not argv:
        return ExitCode.NOT_ENOUGH_ARGS, 'no file path provided'
    if argv[0] == '--help':
        print(yesqa_command.__doc__)
        return ExitCode.OK, ''

    paths = get_paths(Path(fname) for fname in argv)
    fixer = YesQA()
    for path in paths:
        modified = fixer(path=path)
        if modified:
            print(str(path))
    return ExitCode.OK, ''



# built-in
from unittest import mock

# project
from flakehell._patched._checkers import FlakeHellFileChecker


def test_nonexistent_file():
    """Verify that checking non-existent file results in an error."""
    plugin = {
        'plugin_name': 'flake8-example',
        'name': 'something',
        'plugin': FlakeHellFileChecker,
    }
    checks = dict(ast_plugins=[plugin], logical_line_plugins=[], physical_line_plugins=[])
    c = FlakeHellFileChecker(
        filename='foobar.py',
        checks=checks,
        options=None,
    )

    assert c.processor is None
    assert not c.should_process
    assert len(c.results) == 1
    error = c.results[0]
    assert error.error_code == 'E902'


def test_catches_exception_on_invalid_syntax(tmp_path):
    code_path = tmp_path / 'example.py'
    code_path.write_text('I exist!')
    plugin = {
        'name': 'failure',
        'plugin_name': 'failure',
        'parameters': dict(),
        'plugin': mock.MagicMock(side_effect=ValueError),
    }
    options = mock.MagicMock()
    options.safe = False
    checks = dict(ast_plugins=[plugin], logical_line_plugins=[], physical_line_plugins=[])
    fchecker = FlakeHellFileChecker(
        filename=str(code_path),
        checks=checks,
        options=options,
    )
    assert fchecker.should_process is True
    assert fchecker.processor is not None
    fchecker.run_checks()
    assert len(fchecker.results) == 1
    assert fchecker.results[0].error_code == 'E999'
    assert fchecker.results[0].text == 'SyntaxError: invalid syntax'



import ast
from pathlib import Path
from flakehell.parsers import YAMLParser


GIVEN = """
- case: context_ask1
  disable_cache: true
  main: |
    from returns.context import Context
    reveal_type(Context.ask())  # N: ...

- case: context_wrong_cast
  disable_cache: true
  main: |
    from returns.context import RequiresContext
    first: RequiresContext[ValueError, TypeError]  # we can only return type
    second: RequiresContext[Exception, Exception] = first
  out: |
    main:4: error: ...

- case: context_covariant_cast
  disable_cache: true
  main: |
    from returns.context import RequiresContext
    class A(object):
        a = 1
"""


EXPECTED = """
reveal_type = lambda x: x  # noqa
# disable_cache: true
# main: |
from returns.context import Context
reveal_type(Context.ask())  # N: ...

# - case: context_wrong_cast
# disable_cache: true
# main: |
from returns.context import RequiresContext
first: RequiresContext[ValueError, TypeError]  # we can only return type
second: RequiresContext[Exception, Exception] = first
# out: |
# main:4: error: ...

# - case: context_covariant_cast
# disable_cache: true
# main: |
from returns.context import RequiresContext
class A(object):
    a = 1
"""


def test_markdown_parser(tmp_path: Path):
    path = tmp_path / 'test_example.yml'
    path.write_text(GIVEN.lstrip())
    actual = YAMLParser.parse(path=path)
    assert ''.join(actual) == EXPECTED.lstrip()

    # test that the result is a valid python code
    ast.parse(''.join(actual))



# built-in
from pathlib import Path

# external
import pytest

# project
from flakehell._logic import YesQA

# app
from ..utils import chdir


def get_modified(content: str, path: Path) -> str:
    path = path / 'tmp.py'
    path.write_text(content)
    with chdir(path.parent):
        return YesQA().get_modified_file(path=path, original=content)


@pytest.mark.parametrize('content', [
    'print("hello")',
    # don't add not ignored codes
    'err=1',
    # don't touch other codes
    'err=1 # noqa: E225',
    # preserve case
    'err=1 # NoQA: E225',
    # respect tokenization structure
    # 'err="# noqa: E117"',
])
def test_not_modified(content: str, tmp_path: Path):
    assert get_modified(content=content, path=tmp_path) == content


@pytest.mark.parametrize('given, expected', [
    # remove one code
    ('err=1 # noqa: E225, E117', 'err=1 # noqa: E225'),
    ('err=1 # noqa: E117, E225', 'err=1 # noqa: E225'),
    # don't touch text after
    ('err=1 # noqa: E225, E117 # comment', 'err=1 # noqa: E225 # comment'),
    # remove comment when the last code is removed
    ('err=1 # noqa: E117', 'err=1'),
    # case insensitive
    ('err=1 # NoQA: E117', 'err=1'),

    # https://github.com/asottile/yesqa/blob/master/tests/yesqa_test.py
    ('import os  # noqa: F401,X999\n', 'import os  # noqa: F401\n'),
    ('import os  # noqa:F401,X999\n', 'import os  # noqa: F401\n'),
])
def test_modified(given: str, expected: str, tmp_path: Path):
    assert get_modified(content=given, path=tmp_path) == expected



# built-in
import json
import re
from collections import defaultdict
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from typing import Dict, List, Set

# app
from .._constants import NAME, VERSION


CODE = '[a-z]+[0-9]+'
SEP = r'[,\s]+'


class YesQA:
    noqa_file_re = re.compile(r'^# flake8[:=]\s*noqa', re.I)
    noqa_re = re.compile(f'# noqa(: ?{CODE}({SEP}{CODE})*)?', re.I)
    code_re = re.compile(CODE, re.I)

    def get_ignored_codes(self, line: str) -> List[str]:
        match = self.noqa_re.search(line)
        if not match:
            return []
        comment = match.group()
        return self.code_re.findall(comment)

    def remove_noqa(self, line: str) -> str:
        if self.noqa_file_re.match(line):
            return ''
        match = self.noqa_re.search(line)
        if not match:
            return line
        line = line[:match.start()] + line[match.end():]
        return line.rstrip()

    def remove_noqa_code(self, line: str, code: str) -> str:
        match = self.noqa_re.search(line)
        if not match:
            return line
        comment = match.group()
        codes = self.code_re.findall(comment)

        # if it was only one code and we remove it, remove the comment at all
        if codes == [code]:
            return self.remove_noqa(line)

        # remove only one code from the list of codes
        codes = [c for c in codes if c != code]
        new_comment = '# noqa: ' + ', '.join(codes)
        line = line[:match.start()] + new_comment + line[match.end():]
        return line.rstrip()

    def get_errors(self, path: Path, noqa: bool) -> Dict[int, Set[str]]:
        from .._patched import FlakeHellApplication

        app = FlakeHellApplication(program=NAME, version=VERSION)
        output = StringIO()
        cmd = ['--format', 'json', str(path)]
        if not noqa:
            cmd.append('--disable-noqa')
        with redirect_stdout(output):
            app.run(cmd)
        output.seek(0)

        result = defaultdict(set)
        for line in output:
            data = json.loads(line)
            result[data['line']].add(data['code'])
        return dict(result)

    def remove_unused_codes(self, content: str, errors: Dict[int, Set[str]]) -> str:
        result = []
        for line_number, line in enumerate(content.split('\n'), 1):
            ignored_codes = self.get_ignored_codes(line)
            actual_codes = errors.get(line_number, set())
            for code in ignored_codes:
                if code not in actual_codes:
                    line = self.remove_noqa_code(line=line, code=code)
            result.append(line)
        return '\n'.join(result)

    def get_modified_file(self, path: Path, original: str) -> str:
        old_errors = self.get_errors(path=path, noqa=True)
        all_errors = self.get_errors(path=path, noqa=False)
        new_errors = dict()
        for line_number, codes in all_errors.items():
            new_codes = codes - old_errors.get(line_number, set())
            if new_codes:
                new_errors[line_number] = new_codes

        return self.remove_unused_codes(content=original, errors=new_errors)

    def __call__(self, path: Path) -> bool:
        original = path.read_text(encoding='utf-8')
        modified = self.get_modified_file(path=path, original=original)
        if modified == original:
            return False
        path.write_text(modified)
        return True



import ast
from pathlib import Path
from flakehell.parsers import JupyterParser


GIVEN = r"""
{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": ["Example notebook"]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import not_a_package\n",
    "{\"1\": 1}"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "One more line of text"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": { },
   "outputs": [],
   "source": [
    "{\"2\": 1}\n",
    "{\"2\": 2}"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
"""


EXPECTED = """
# In [1]:
import not_a_package
{"1": 1}

# In [2]:
{"2": 1}
{"2": 2}
"""


def test_markdown_parser(tmp_path: Path):
    path = tmp_path / 'example.md'
    path.write_text(GIVEN)
    actual = JupyterParser.parse(path=path)
    assert ''.join(actual) == EXPECTED

    # test that the result is a valid python code
    ast.parse(''.join(actual))



import ast
from pathlib import Path
from flakehell.parsers import RSTParser


GIVEN = """
I think I saw you in my sleep, Python

.. sourcecode:: pycon

    >>> # I think I saw you in my dreams
    >>> You = were(stitching="up the seams")
    'You were stitching up the seams'

On every broken promise that your body couldn't keep

.. code:: python

  from returns.result import Result, Success
  def may_fail(user_id: int) -> Result[float, str]:
      print('look at me!')

I think I saw you in my sleep
"""


EXPECTED = """
# I think I saw you in my sleep, Python

# .. sourcecode:: pycon

# I think I saw you in my dreams
You = were(stitching="up the seams")
# 'You were stitching up the seams'

# On every broken promise that your body c

# .. code:: python

from returns.result import Result, Success
def may_fail(user_id: int) -> Result[float, str]:
    print('look at me!')

# I think I saw you in my sleep
"""


def test_markdown_parser(tmp_path: Path):
    path = tmp_path / 'example.md'
    path.write_text(GIVEN)
    actual = RSTParser.parse(path=path)
    assert ''.join(actual) == EXPECTED

    # test that the result is a valid python code
    ast.parse(''.join(actual))



# built-in
import subprocess
import sys
from io import BytesIO
from unittest.mock import patch, Mock
from pathlib import Path
from textwrap import dedent

# external
import pytest

# project
from flakehell._cli import main

# app
from .utils import chdir


def test_flake8helled_file():
    """Baseline behavior, when an actual filename is passed."""
    cmd = [
        sys.executable,
        '-c',
        'import sys; from flakehell import flake8_entrypoint; sys.exit(flake8_entrypoint())',
        __file__,
    ]
    result = subprocess.run(cmd)
    assert result.returncode == 0


def test_flake8helled_stdin():
    """Problematic behavior from issue #44, `-` is passed as filename, together with --stdin-display-name."""
    source_file = open(__file__, 'r')
    cmd = [
        sys.executable,
        '-c',
        'import sys; from flakehell import flake8_entrypoint; sys.exit(flake8_entrypoint())',
        '--stdin-display-name',
        __file__,
        # '-' is not an existing filename, so snapshot cannot create a hexdigest of its content
        # but otherwise it's fine for flake8 which knows to read stdin instead
        '-',
    ]
    result = subprocess.run(cmd, stdin=source_file)
    assert result.returncode == 0


@pytest.mark.parametrize('flag', [
    '--help',
    'help',
    'commands',
])
def test_help(flag, capsys):
    result = main([flag])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''

    for name in ('baseline', 'code', 'codes', 'lint', 'missed', 'plugins'):
        assert name in captured.out


def test_version(capsys):
    result = main(['--version'])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    assert 'FlakeHell' in captured.out
    assert 'Flake8' in captured.out


def test_lint_help(capsys):
    result = main(['lint', '--help'])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''

    # flake8 options
    assert '-h, --help' in captured.out
    assert '--builtins' in captured.out
    assert '--isort-show-traceback' in captured.out

    # ignored flake8 options
    assert '--per-file-ignores' not in captured.out
    assert '--enable-extensions' not in captured.out

    # flakehell options
    assert '--baseline' in captured.out


def test_exceptions(capsys, tmp_path: Path):
    text = """
    [tool.flakehell.plugins]
    pyflakes = ["+*"]

    [tool.flakehell.exceptions."tests/"]
    pyflakes = ["-F401"]
    """
    (tmp_path / 'pyproject.toml').write_text(dedent(text))
    (tmp_path / 'example.py').write_text('import sys\na')
    (tmp_path / 'tests').mkdir()
    (tmp_path / 'tests' / 'test_example.py').write_text('import sys\na')
    with chdir(tmp_path):
        result = main(['lint', '--format', 'default'])
    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    exp = """
    ./example.py:1:1: F401 'sys' imported but unused
    ./example.py:2:1: F821 undefined name 'a'
    ./tests/test_example.py:2:1: F821 undefined name 'a'
    """
    assert captured.out.strip() == dedent(exp).strip()


@patch('sys.stdin.buffer', BytesIO(b'import sys\na\n'))
@patch('sys.stdin', Mock())
def test_exceptions_stdin(capsys, tmp_path: Path):
    # write config
    text = """
    [tool.flakehell.plugins]
    pyflakes = ["+*"]

    [tool.flakehell.exceptions."example.py"]
    pyflakes = ["-F401"]
    """
    (tmp_path / 'pyproject.toml').write_text(dedent(text))

    # make fake `stdin` with source file content
    # code_path = tmp_path / 'example.py'
    # code_path.write_text('import sys\na')

    # call the command with passing matching `--stdin-display-name`
    cmd = ['lint', '--format', 'default', '--stdin-display-name', 'example.py', '--', '-']
    with chdir(tmp_path):
        # with code_path.open('r') as stream:
        result = main(cmd)

    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    assert captured.out.strip() == "example.py:2:1: F821 undefined name 'a'"


def test_baseline(capsys, tmp_path: Path):
    code_path = tmp_path / 'example.py'
    code_path.write_text('a\nb\n')
    with chdir(tmp_path):
        result = main(['baseline', str(code_path)])
    assert result == (0, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    hashes = captured.out.strip().split()
    assert len(hashes) == 2

    line_path = tmp_path / 'baseline.txt'
    line_path.write_text(hashes[0])
    with chdir(tmp_path):
        result = main([
            'lint',
            '--baseline', str(line_path),
            '--format', 'default',
            str(code_path),
        ])
    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    assert captured.out.strip() == "{}:2:1: F821 undefined name 'b'".format(str(code_path))


def test_ignore_file_by_top_level_noqa(capsys, tmp_path: Path):
    (tmp_path / 'example1.py').write_text('import sys\n')
    (tmp_path / 'example2.py').write_text('# flake8: noqa\nimport sys\n')
    with chdir(tmp_path):
        result = main(['lint', '--format', 'default'])
    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    exp = "./example1.py:1:1: F401 'sys' imported but unused"
    assert captured.out.strip() == exp


def test_exclude_file(capsys, tmp_path: Path):
    (tmp_path / 'checked.py').write_text('import sys\n')
    (tmp_path / 'ignored').mkdir()
    (tmp_path / 'ignored' / 'first.py').write_text('import sys\n')
    (tmp_path / 'ignored' / 'second.py').write_text('invalid syntax!')
    with chdir(tmp_path):
        result = main(['lint', '--format', 'default', '--exclude', 'ignored'])
    assert result == (1, '')
    captured = capsys.readouterr()
    assert captured.err == ''
    exp = """
    ./checked.py:1:1: F401 'sys' imported but unused
    """
    assert captured.out.strip() == dedent(exp).strip()



# external
import pytest

# project
from flakehell._constants import KNOWN_PLUGINS
from flakehell._logic import extract, get_installed, _extractors
from flakehell._patched import FlakeHellApplication


@pytest.mark.parametrize('plugin_name', KNOWN_PLUGINS)
def test_smoke_extract(plugin_name):
    codes = extract(plugin_name)
    assert codes

    for code, msg in codes.items():
        assert type(code) is str, 'bad code type'
        assert type(msg) is str, 'bad message type'

        # that's not exactly true but all plugins follow this convention
        assert code[0].isalpha(), 'code must start from letter'
        assert code[0].isupper(), 'code must be uppercase'


@pytest.mark.parametrize('plugin_name', KNOWN_PLUGINS)
def test_smoke_prefixes(plugin_name):
    app = FlakeHellApplication(program='test', version='1.0.0')
    plugins = {plugin['name']: plugin for plugin in get_installed(app=app)}
    plugin = plugins[plugin_name]

    codes = extract(plugin_name)
    for code in codes:
        print(plugin_name, code, plugin['codes'])
        assert code.startswith(tuple(plugin['codes']))


@pytest.mark.parametrize('plugin_name', KNOWN_PLUGINS)
def test_no_custom_extractor_needed(plugin_name):
    extractor = getattr(_extractors, 'extract_' + plugin_name, None)
    if extractor is None:
        return
    custom_codes = extractor()
    default_codes = _extractors.extract_default(plugin_name)
    assert default_codes != custom_codes


def test_no_duplicate_links():
    all_urls = sorted(KNOWN_PLUGINS.values())
    unique_urls = sorted(set(all_urls))
    assert all_urls == unique_urls

    for url in all_urls:
        assert url.startswith('https://')



# external
from flake8 import __version__ as flake8_version
from termcolor import colored

# app
from .._constants import ExitCode
from .._types import CommandResult
from .._version import __version__ as flakehell_version


def version_command(argv) -> CommandResult:
    """Show FlakeHell version.
    """
    print('FlakeHell', colored(flakehell_version, 'green'))
    print('Flake8   ', colored(flake8_version, 'green'))
    print('For plugins versions use', colored('flakehell plugins', 'green'))
    return ExitCode.OK, ''



import ast
from pathlib import Path
from flakehell.parsers import MarkdownParser


GIVEN = """
# Example

This is example of markdown file

```python
a = 1
print('oh hi mark')
```

Emphasized:

```python hl_lines="1"
emphasized_imaginary_function()
```

PyCon:

```pycon
>>> print("Hello")
'Hello'
>>> banana = "banana"
>>> for character in banana:
...     print(characterr)

```

1. Indented:
    ```python
    if 'look at me!':
        print('i am indented!')
    ```
"""


EXPECTED = """
# # Example

# This is example of markdown file

# ```python
a = 1
print('oh hi mark')
# ```

# Emphasized:

# ```python hl_lines="1"
emphasized_imaginary_function()
# ```

# PyCon:

# ```pycon
print("Hello")
# 'Hello'
banana = "banana"
for character in banana:
    print(characterr)

# ```

# 1. Indented:
# ```python
if 'look at me!':
    print('i am indented!')
# ```
"""


def test_markdown_parser(tmp_path: Path):
    path = tmp_path / 'example.md'
    path.write_text(GIVEN)
    actual = MarkdownParser.parse(path=path)
    assert ''.join(actual) == EXPECTED

    # test that the result is a valid python code
    ast.parse(''.join(actual))
